from starlette.responses import Response
import os
import mimetypes


class EmptyWSGI:
    def __init__(self):
        ...


class StaticFile:
    def __init__(self, directory):
        self.directory = directory
    
    def __call__(self, request, filename: str):
        headers = {}
        path = os.path.relpath(filename, os.getcwd())
        
        with open(f"{self.directory}/{path}", "rb") as f:
            data = f.read()
        
        mimetype = mimetypes.guess_type(filename)
        
        headers.update(
            {
                "Content-Type": mimetype
            }
        )
        
        return Response(data, media_type=mimetype[0])





from arc.defaults import DefaultExceptionHandler
from arc.staticfiles import StaticFile
from jinja2.loaders import FileSystemLoader
from jinja2 import Environment, FileSystemLoader
import os


class Collection:
    def __init__(self, templates_dir="templates", static_dir="static", exception_handler=None, host="127.0.0.1", port=5000):
        self.routes = {}

        # The host
        self.host = host

        # The port
        self.port = port

        self.templates_dir = templates_dir

        self.templates_env = Environment(
            loader=FileSystemLoader(os.path.abspath(self.templates_dir)))

        if exception_handler is None:
            self.exception_handler = DefaultExceptionHandler(self)
        else:
            self.exception_handler = exception_handler()

        self.static_app = StaticFile(static_dir)

        self.add_route("/static/{filename}", self.static_app)

    def route(self, path):

        assert path not in self.routes, f"Route {path} already exists"

        def wrapper(handler):
            self.add_route(path, handler)
            return handler

        return wrapper

    def add_route(self, path, handler):
        assert path not in self.routes, f"Route {path} already exists"

        self.routes[path] = handler

    def change_template_env(self, templates_dir):
        self.templates_env = Environment(
            loader=FileSystemLoader(os.path.abspath(templates_dir)))
        self.templates_dir = templates_dir



from arc.app import *
from arc.errors import *
from arc.middleware import *
from arc.defaults import *
from arc.sessions import *
from arc.collections import *
from arc.cli import *
from arc.templates import *



from arc.middleware import Middleware
from arc.defaults import LoggingMiddleware, DefaultExceptionHandler
from arc.staticfiles import StaticFile
from arc.logger import Logger
from jinja2.loaders import FileSystemLoader
import uvicorn
from parse import parse
from jinja2 import Environment, FileSystemLoader
import inspect
import os
import traceback


class App:
    """
    The base App object. This object creates a WSGI application, and is used as the base object for creating routes, views, 
    templating, and more. Basic usage is as the following:: 

    from arc import App

    app = App()

    @app.route("/")
    def index(req, res):
        res.text = 'Hello, World'

    if __name__ == "__main__":
        app.run()

    The app has several parameters.

    ..templates_dir ::
        This parameter is used to specify the directory that jinja loads its HTML templates from. Its defaulted to 
        the 'templates' directory, but you can change it as the following::

        app = App(templates_dir="public")

    ..static_dir ::
        This parameter is used to specify the directory that static files are served from. Its defaulted to the
        'static' folder. You can change it as the following::

        app = App(static_dir="css")

    ..exception_handler ::
        This parameter is used to specify a custom exception handler for the app to use instead of the default
        one. It uses the base AppException class. It's defaulted to None, which causes the app to use the
        default exception handler. It can be changed as the following::

        app = App(exception_handler=CustomExceptionHandler)

    ..host ::
        This parameter is used to specify the host that the app will run on. Its defaulted 
        to '127.0.0.1'. It can be changed as the following::

        app = App(host="181.177.121.773")

    ..port ::
        This parameter is used to specify which port the app will listen on. Its defaulted to 
        port `5000`. You can change it as the following::

        app = App(port=8000)

    ..logging
        This parameter specifies whether the apps default logger should be activated. It can
        be deactivated as the following::

        app = App(logging=False)

    ..logger
        If the default logger is enabled, it specifies the logger the app should use.
        It can be changed as the following::

        app = App(logger=MyCustomLogger())
    """

    def __init__(self, templates_dir: str = "templates", static_dir: str = "static", exception_handler=None, host: str = "127.0.0.1", port=5000, logging=True, logger=Logger()):
        self.routes = {}

        # The host
        self.host = host

        # The port
        self.port = port

        self.templates_dir = templates_dir

        if exception_handler is None:
            self.exception_handler = DefaultExceptionHandler(self)
        else:
            self.exception_handler = exception_handler()

        self.middleware = Middleware(self)

        if logging:
            self.add_middleware(LoggingMiddleware)

        self.static_app = StaticFile(static_dir)

        self.collections = []

        self.add_route("/static/{filename}", self.static_app)

        self.logger = logger

    async def __call__(self, scope, receive, send):
        await self.middleware(scope, receive, send)

    def handle_request(self, request):
        handler, kwargs = self.find_handler(request_path=request.url.path)

        try:
            if handler is not None:
                if inspect.isclass(handler):
                    handler = getattr(handler(), request.method.lower(), None)
                    if handler is None:
                        raise AttributeError(
                            "Method not allowed", request.method)

                response = handler(request, **kwargs)
            else:
                return self.default_response(request)

        except Exception as e:
            error = traceback.format_exc()
            response = self.exception_handler.handle_error(request, error)

        return response

    def default_response(self, request):
        return self.exception_handler.handle_404(request)

    def find_handler(self, request_path):
        for path, handler in self.routes.items():
            parse_result = parse(path, request_path)
            if parse_result is not None:
                return handler, parse_result.named

        return None, None

    def route(self, path):

        assert path not in self.routes, f"Route {path} already exists"

        def wrapper(handler):
            self.add_route(path, handler)
            return handler

        return wrapper

    def add_route(self, path, handler):
        assert path not in self.routes, f"Route {path} already exists"

        self.routes[path] = handler

    def add_middleware(self, middleware_cls):
        self.middleware.add(middleware_cls)

    def change_template_env(self, templates_dir):
        self.templates_env = Environment(
            loader=FileSystemLoader(os.path.abspath(templates_dir)))
        self.templates_dir = templates_dir

    def register_collection(self, collection):
        assert collection not in self.collections, f"Collection {collection} already registered"
        collection.set_wsgi(self.wsgi_app)
        for path in collection.routes.keys():
            assert path not in self.routes, f"Route {path} already exists"

            self.routes[path] = collection.routes[path]

        self.routes.update(collection.routes)
        self.collections.append(collection)

    def run(self):
        self.logger.log(f"Running on http://{self.host}:{self.port}", "info")
        self.logger.log(f"Press CTRL + C to stop", "info")
        uvicorn.run(self, host="127.0.0.1", port=5000,
                    log_level="critical")
        self.logger.log("Exiting Application", "critical")



from jinja2 import Environment, FileSystemLoader
from starlette.responses import Response
import jinja2
import typing


class Template:
    def __init__(self, directory: str) -> None:
        self.env = self.get_env(directory)

    def get_env(self, directory: str) -> "Environment":
        @jinja2.contextfunction
        def url_for(context: dict, name: str, **path_params: typing.Any) -> str:
            request = context["request"]
            return request.url_for(name, **path_params)

        loader = FileSystemLoader(directory)
        env = jinja2.Environment(loader=loader, autoescape=True)

        env.globals["url_for"] = url_for

        return env

    def get_template(self, name: str) -> "jinja2.Template":
        return self.env.get_template(name)

    def __call__(self, name: str, context: dict, status_code: int = 200, headers: dict = None, media_type: str = None) -> Response:
        if "request" not in context:
            raise ValueError('Context must include a "request" key')

        template = self.env.get_template(name)

        return Response(
            template.render(context),
            status_code=status_code,
            headers=headers,
            media_type=media_type
        )



import logging
from logging.config import dictConfig
import json


class Logger:
    def __init__(self, config_dict: dict = None):
        if config_dict is not None:
            self.config_dict = config_dict
        else:
            self.config_dict = dict(
                version = 1,
                formatters={
                    'f': {'format':
                          '%(levelname)s: %(message)s'}
                },
                handlers={
                    'h': {'class': 'logging.StreamHandler',
                          'formatter': 'f',
                          'level': logging.DEBUG}
                },
                root={
                    'handlers': ['h'],
                    'level': logging.DEBUG,
                },
            )

        dictConfig(self.config_dict)

        self.logger = logging.getLogger()

    def log(self, message: str, level: str) -> None:
        levels = {
            "info": self.logger.info,
            "debug": self.logger.debug,
            "warning": self.logger.warning,
            "error": self.logger.error,
            "critical": self.logger.critical
        }

        levels[level](message)


def config_from_json(path: str) -> dict:
    with open(path, "r") as f:
        config_dict = json.load(f)

    return config_dict



from starlette.requests import Request


class Middleware:
    def __init__(self, app):
        self.app = app
        self.test = "YAY"

    async def __call__(self, scope, receive, send):
        request = Request(scope=scope, receive=receive)

        response = self.app.handle_request(request)

        await response(scope, receive, send)

    def add(self, middleware_cls):
        self.app = middleware_cls(self.app)

    def process_request(self, req):
        pass

    def process_response(self, req, res):
        pass

    def handle_request(self, request):
        self.process_request(request)

        response = self.app.handle_request(request)
        self.process_response(request, response)

        return response



# import argparse
# import os
# import sys


# parser = argparse.ArgumentParser(
#     description="CLI for interacting with the Arc Web Framework", prog="arc")


# parser.add_argument(
#     "Data",
#     metavar="data",
#     type=str,
#     help="test"
# )

# args = parser.parse_args()

# data = args.Data

# print("Repeated" + data)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="arcframework",  # Replace with your own username
    version="1.9.5",
    author="Rayyan Cyclegar",
    author_email="aboominister@gmail.com",
    description="A Python micro web framework for creating dynamic websites.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AbooMinister25/arc",
    packages=["arc", ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        "uvicorn", "parse", "jinja2", "starlette"
    ],
    include_package_data=True,
    package_data={
        "arc": ["selfpages/*.html", "static/*.css"],
    }
)





class Session:
    def __init__(self, app, lifetime="session"):
        self.app = app
        self._dict = {}
        self.lifetime = lifetime

    def __getitem__(self, key):
        return self._dict[key]

    def make_session(self, key, value, resp):
        if self.lifetime == "session":
            self.app.set_cookie(key, value, resp)
        else:
            self.app.set_cookie(key, value, resp, lifetime=self.lifetime)

        self._dict[key] = value



from arc.middleware import Middleware
from arc.errors import AppException
from arc.templates import Template
import os


class LoggingMiddleware(Middleware):
    def process_request(self, req):
        self.app.logger.log(f"\n[REQUEST][{req.method}] {req.url}", "info")

    def process_response(self, req, res):
        self.app.logger.log(f"\n[RESPONSE] {req.url}", "info")


class DefaultExceptionHandler:
    def __init__(self, app):
        self.app = app
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.path = os.path.join(self.dir_path, "selfpages")
        self.template = Template(self.path)

    def handle_error(self, request, error: str):
        self.app.logger.log(error, "error")
        return self.template("error-500.html", context={"request": request, "error": error})

    def handle_404(self, request):
        self.app.logger.log(f"{request.url.path} not found", "error")
        return self.template("error-404.html", context={"request": request})
